# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import raise_if_error
from nbxmpp.modules.bookmarks.util import parse_bookmarks
from nbxmpp.modules.bookmarks.util import build_storage_node


BOOKMARK_OPTIONS = {
    'pubsub#persist_items': 'true',
    'pubsub#access_model': 'whitelist',
}


class PEPBookmarks(BaseModule):

    _depends = {
        'publish': 'PubSub',
        'request_items': 'PubSub',
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_bookmarks,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
        ]

    def _process_pubsub_bookmarks(self, _client, stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.BOOKMARKS:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            bookmarks = parse_bookmarks(item, self._log)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if not bookmarks:
            self._log.info('Bookmarks removed')
            return

        pubsub_event = properties.pubsub_event._replace(data=bookmarks)
        self._log.info('Received bookmarks from: %s', properties.jid)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def request_bookmarks(self):
        _task = yield

        items = yield self.request_items(Namespace.BOOKMARKS, max_items=1)
        raise_if_error(items)

        if not items:
            yield []

        bookmarks = parse_bookmarks(items[0], self._log)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        yield bookmarks

    @iq_request_task
    def store_bookmarks(self, bookmarks):
        _task = yield

        self._log.info('Store Bookmarks')

        self.publish(Namespace.BOOKMARKS,
                     build_storage_node(bookmarks),
                     id_='current',
                     options=BOOKMARK_OPTIONS,
                     force_node_options=True)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import logging

from nbxmpp.util import LogAdapter

if TYPE_CHECKING:
    from nbxmpp.client import Client


class BaseModule:

    _depends: dict[str, str] = {}

    def __init__(self, client: Client):
        logger_name = 'nbxmpp.m.%s' % self.__class__.__name__.lower()
        self._log = LogAdapter(logging.getLogger(logger_name),
                               {'context': client.log_context})

    def __getattr__(self, name: str) -> Any:
        if name not in self._depends:
            raise AttributeError('Unknown method: %s' % name)

        module = self._client.get_module(self._depends[name])
        return getattr(module, name)



#!/usr/bin/env python3

import re
import argparse
from pathlib import Path
import subprocess

REPO_DIR = Path(__file__).resolve().parent.parent


INIT = REPO_DIR / 'nbxmpp' / '__init__.py'
CHANGELOG = REPO_DIR / 'ChangeLog'

VERSION_RX = r'\d+\.\d+\.\d+'


def get_current_version() -> str:
    content = INIT.read_text(encoding='utf8')
    match = re.search(VERSION_RX, content)
    if match is None:
        exit('Unable to find current version')
    return match[0]


def bump_version(current_version: str, new_version: str) -> None:
    content = INIT.read_text(encoding='utf8')
    content = content.replace(current_version, new_version, 1)
    INIT.write_text(content, encoding='utf8')


def make_changelog(new_version: str) -> None:

    cmd = [
        'git-chglog',
        '--next-tag',
        new_version
    ]

    result = subprocess.run(cmd,
                            cwd=REPO_DIR,
                            text=True,
                            check=True,
                            capture_output=True)

    changes = result.stdout
    changes = changes.removeprefix('\n')

    current_changelog = CHANGELOG.read_text()

    with CHANGELOG.open('w') as f:
        f.write(changes + current_changelog)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Bump Version')
    parser.add_argument('version', help='The new version, e.g. 1.5.0')
    args = parser.parse_args()

    current_version = get_current_version()

    bump_version(current_version, args.version)
    make_changelog(args.version)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.structs import StanzaHandler
from nbxmpp.const import PresenceType
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.task import iq_request_task


class Nickname(BaseModule):

    _depends = {
        'publish': 'PubSub'
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_nickname,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
            StanzaHandler(name='message',
                          callback=self._process_nickname,
                          ns=Namespace.NICK,
                          priority=40),
            StanzaHandler(name='presence',
                          callback=self._process_nickname,
                          ns=Namespace.NICK,
                          priority=40),
        ]

    def _process_nickname(self, _client, stanza, properties):
        if stanza.getName() == 'message':
            properties.nickname = self._parse_nickname(stanza)

        elif stanza.getName() == 'presence':
            # the nickname MUST NOT be included in presence broadcasts
            # (i.e., <presence/> stanzas with no 'type' attribute or
            # of type "unavailable").
            if properties.type in (PresenceType.AVAILABLE,
                                   PresenceType.UNAVAILABLE):
                return
            properties.nickname = self._parse_nickname(stanza)

    def _process_pubsub_nickname(self, _client, _stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.NICK:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        nick = self._parse_nickname(item)
        if nick is None:
            self._log.info('Received nickname: %s - nickname removed',
                           properties.jid)
            return

        self._log.info('Received nickname: %s - %s', properties.jid, nick)
        properties.pubsub_event = properties.pubsub_event._replace(data=nick)

    @staticmethod
    def _parse_nickname(stanza):
        nickname = stanza.getTag('nick', namespace=Namespace.NICK)
        if nickname is None:
            return None
        return nickname.getData() or None

    @iq_request_task
    def set_nickname(self, nickname, public=False):
        task = yield

        access_model = 'open' if public else 'presence'

        options = {
            'pubsub#persist_items': 'true',
            'pubsub#access_model': access_model,
        }

        item = Node('nick', {'xmlns': Namespace.NICK})
        if nickname is not None:
            item.addData(nickname)

        result = yield self.publish(Namespace.NICK,
                                    item,
                                    id_='current',
                                    options=options,
                                    force_node_options=True)

        yield finalize(task, result)

    @iq_request_task
    def set_access_model(self, public):
        task = yield

        access_model = 'open' if public else 'presence'

        result = yield self._client.get_module('PubSub').set_access_model(
            Namespace.NICK, access_model)

        yield finalize(task, result)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import LocationData
from nbxmpp.const import LOCATION_DATA
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.task import iq_request_task


class Location(BaseModule):

    _depends = {
        'publish': 'PubSub'
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_location,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
        ]

    def _process_pubsub_location(self, _client, _stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.LOCATION:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        location_node = item.getTag('geoloc', namespace=Namespace.LOCATION)
        if not location_node.getChildren():
            self._log.info('Received location: %s - no location set',
                           properties.jid)
            return

        location_dict = {}
        for node in LOCATION_DATA:
            location_dict[node] = location_node.getTagData(node)
        data = LocationData(**location_dict)
        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info('Received location: %s - %s', properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_location(self, data):
        task = yield

        item = Node('geoloc', {'xmlns': Namespace.LOCATION})
        if data is not None:
            data = data._asdict()
            for tag, value in data.items():
                if value is not None:
                    item.addChild(tag, payload=value)

        result = yield self.publish(Namespace.LOCATION, item, id_='current')

        yield finalize(task, result)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.modules.base import BaseModule


class PGPLegacy(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pgplegacy_message,
                          ns=Namespace.ENCRYPTED,
                          priority=7),
            StanzaHandler(name='presence',
                          callback=self._process_signed,
                          ns=Namespace.SIGNED,
                          priority=15)
        ]

    @staticmethod
    def _process_signed(_client, stanza, properties):
        signed = stanza.getTag('x', namespace=Namespace.SIGNED)
        if signed is None:
            return

        properties.signed = signed.getData()

    def _process_pgplegacy_message(self, _client, stanza, properties):
        pgplegacy = stanza.getTag('x', namespace=Namespace.ENCRYPTED)
        if pgplegacy is None:
            self._log.warning('No x node found')
            self._log.warning(stanza)
            return

        data = pgplegacy.getData()
        if not data:
            self._log.warning('No data in x node found')
            self._log.warning(stanza)
            return

        self._log.info('Encrypted message received')
        properties.pgp_legacy = data



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.protocol import JID
from nbxmpp.structs import AnnotationNote
from nbxmpp.errors import StanzaError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.modules.date_and_time import parse_datetime


class Annotations(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @property
    def domain(self):
        return self._client.get_bound_jid().domain

    @iq_request_task
    def request_annotations(self):
        _task = yield

        response = yield _make_request()
        if response.isError():
            raise StanzaError(response)

        query = response.getQuery()
        storage = query.getTag('storage', namespace=Namespace.ROSTERNOTES)
        if storage is None:
            raise MalformedStanzaError('storage node missing', response)

        notes = []
        for note in storage.getTags('note'):
            try:
                jid = JID.from_string(note.getAttr('jid'))
            except Exception as error:
                self._log.warning('Invalid JID: %s, %s',
                                  note.getAttr('jid'), error)
                continue

            cdate = note.getAttr('cdate')
            if cdate is not None:
                cdate = parse_datetime(cdate, epoch=True)

            mdate = note.getAttr('mdate')
            if mdate is not None:
                mdate = parse_datetime(mdate, epoch=True)

            data = note.getData()
            notes.append(AnnotationNote(jid=jid, cdate=cdate,
                                        mdate=mdate, data=data))

        self._log.info('Received annotations from %s:', self.domain)
        for note in notes:
            self._log.info(note)
        yield notes

    @iq_request_task
    def set_annotations(self, notes):
        _task = yield

        for note in notes:
            self._log.info(note)

        response = yield _make_set_request(notes)
        yield process_response(response)


def _make_request():
    payload = Node('storage', attrs={'xmlns': Namespace.ROSTERNOTES})
    return Iq(typ='get', queryNS=Namespace.PRIVATE, payload=payload)

def _make_set_request(notes):
    storage = Node('storage', attrs={'xmlns': Namespace.ROSTERNOTES})
    for note in notes:
        node = Node('note', attrs={'jid': note.jid})
        node.setData(note.data)
        if note.cdate is not None:
            node.setAttr('cdate', note.cdate)
        if note.mdate is not None:
            node.setAttr('mdate', note.mdate)
        storage.addChild(node=node)
    return Iq(typ='set', queryNS=Namespace.PRIVATE, payload=storage)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
from typing import Optional

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Protocol


def is_error(error):
    return isinstance(error, BaseError)


class BaseError(Exception):
    def __init__(self, is_fatal: bool = False) -> None:
        self.is_fatal = is_fatal
        self.text = ''

    def __str__(self) -> str:
        return self.text

    def get_text(self, _pref_lang: Optional[str] = None) -> str:
        return self.text


class StanzaError(BaseError):

    log_level = logging.INFO
    app_namespace = None

    def __init__(self, stanza: Protocol):
        BaseError.__init__(self)
        self.stanza = stanza
        self._stanza_name = stanza.getName()
        self._error_node = stanza.getTag('error')
        self.condition: Optional[str] = stanza.getError()
        self.condition_data = self._error_node.getTagData(self.condition)
        self.app_condition = self._get_app_condition()
        self.type = stanza.getErrorType()
        self.jid = stanza.getFrom()
        self.id = stanza.getID()
        self._text: dict[str, str] = {}

        text_elements = self._error_node.getTags('text',
                                                 namespace=Namespace.STANZAS)
        for element in text_elements:
            lang = element.getXmlLang()
            text = element.getData()
            self._text[lang] = text

    def _get_app_condition(self) -> Optional[str]:
        if self.app_namespace is None:
            return None

        for node in self._error_node.getChildren():
            if node.getNamespace() == self.app_namespace:
                return node.getName()
        return None

    def get_text(self, pref_lang: Optional[str] = None) -> str:
        if pref_lang is not None:
            text = self._text.get(pref_lang)
            if text is not None:
                return text

        if self._text:
            text = self._text.get('en')
            if text is not None:
                return text

            text = self._text.get(None)
            if text is not None:
                return text
            return self._text.popitem()[1]
        return ''

    def set_text(self, lang: str, text: str) -> None:
        self._text[lang] = text

    def __str__(self) -> str:
        condition = self.condition
        if self.app_condition is not None:
            condition = '%s (%s)' % (self.condition, self.app_condition)
        text = self.get_text('en') or ''
        if text:
            text = ' - %s' % text
        return 'Error from %s: %s%s' % (self.jid, condition, text)


class PubSubStanzaError(StanzaError):

    app_namespace = Namespace.PUBSUB_ERROR


class HTTPUploadStanzaError(StanzaError):

    app_namespace = Namespace.HTTPUPLOAD_0

    def get_max_file_size(self):
        if self.app_condition != 'file-too-large':
            return None

        node = self._error_node.getTag(self.app_condition)
        try:
            return float(node.getTagData('max-file-size'))
        except Exception:
            return None

    def get_retry_date(self):
        if self.app_condition != 'retry':
            return None
        return self._error_node.getTagAttr('stamp')


class MalformedStanzaError(BaseError):

    log_level = logging.WARNING

    def __init__(self, text, stanza, is_fatal=True):
        BaseError.__init__(self, is_fatal=is_fatal)
        self.stanza = stanza
        self.text = str(text)


class CancelledError(BaseError):

    log_level = logging.INFO

    def __init__(self):
        BaseError.__init__(self, is_fatal=True)
        self.text = 'Task has been cancelled'


class TimeoutStanzaError(BaseError):

    log_level = logging.INFO

    def __init__(self):
        BaseError.__init__(self)
        self.text = 'Timeout reached'


class RegisterStanzaError(StanzaError):
    def __init__(self, stanza, data):
        StanzaError.__init__(self, stanza)
        self._data = data

    def get_data(self):
        return self._data


class ChangePasswordStanzaError(StanzaError):
    def __init__(self, stanza, form):
        StanzaError.__init__(self, stanza)
        self._form = form

    def get_form(self):
        return self._form



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.structs import HTTPUploadData
from nbxmpp.errors import HTTPUploadStanzaError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule


ALLOWED_HEADERS = ['Authorization', 'Cookie', 'Expires']


class HTTPUpload(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_slot(self, jid, filename, size, content_type):
        _task = yield

        response = yield _make_request(jid, filename, size, content_type)
        if response.isError():
            raise HTTPUploadStanzaError(response)

        slot = response.getTag('slot', namespace=Namespace.HTTPUPLOAD_0)
        if slot is None:
            raise MalformedStanzaError('slot node missing', response)

        put_uri = slot.getTagAttr('put', 'url')
        if put_uri is None:
            raise MalformedStanzaError('put uri missing', response)

        get_uri = slot.getTagAttr('get', 'url')
        if get_uri is None:
            raise MalformedStanzaError('get uri missing', response)

        headers = {}
        for header in slot.getTag('put').getTags('header'):
            name = header.getAttr('name')
            if name not in ALLOWED_HEADERS:
                raise MalformedStanzaError(
                    'not allowed header found: %s' % name, response)

            data = header.getData()
            if '\n' in data:
                raise MalformedStanzaError(
                    'newline in header data found', response)

            headers[name] = data

        yield HTTPUploadData(put_uri=put_uri,
                             get_uri=get_uri,
                             headers=headers)


def _make_request(jid, filename, size, content_type):
    iq = Iq(typ='get', to=jid)
    attr = {'filename': filename,
            'size': size,
            'content-type': content_type}
    iq.setTag(name="request",
              namespace=Namespace.HTTPUPLOAD_0,
              attrs=attr)
    return iq



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import PubSubEventData
from nbxmpp.modules.user_avatar import AvatarMetaData
from nbxmpp.modules.user_avatar import AvatarInfo

class AvatarTest(StanzaHandlerTest):

    def test_avatar_parsing(self):
        def _on_message(_con, _stanza, properties):
            info = AvatarInfo(bytes='12345',
                              height='64',
                              width='64',
                              id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f',
                              type='image/png',
                              url='http://avatars.example.org/happy.gif')

            metadata = AvatarMetaData(infos=[info])
            pubsub_event = PubSubEventData(
                node='urn:xmpp:avatar:metadata',
                id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f',
                item=None,
                data=metadata,
                deleted=False,
                retracted=False,
                purged=False)

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = '''
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='urn:xmpp:avatar:metadata'>
                        <item id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f'>
                            <metadata xmlns='urn:xmpp:avatar:metadata'>
                                <info bytes='12345'
                                      height='64'
                                      id='111f4b3c50d7b0df729d299bc6f8e9ef9066971f'
                                      type='image/png'
                                      width='64'
                                      url='http://avatars.example.org/happy.gif'/>
                            </metadata>
                        </item>
                    </items>
                </event>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT))

        self.dispatcher.process_data(event)



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import PubSubEventData


class PubsubTest(StanzaHandlerTest):

    def test_purge_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node='princely_musings',
                id=None,
                item=None,
                data=None,
                deleted=False,
                retracted=False,
                purged=True)

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = '''
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <purge node='princely_musings'/>
                </event>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16))

        self.dispatcher.process_data(event)

    def test_delete_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node='princely_musings',
                id=None,
                item=None,
                data=None,
                deleted=True,
                retracted=False,
                purged=False)

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = '''
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <delete node='princely_musings'>
                    <redirect uri='xmpp:hamlet@denmark.lit?;node=blog'/>
                </delete>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16))

        self.dispatcher.process_data(event)


    def test_retracted_event(self):
        def _on_message(_con, _stanza, properties):

            pubsub_event = PubSubEventData(
                node='princely_musings',
                id='ae890ac52d0df67ed7cfdf51b644e901',
                item=None,
                data=None,
                deleted=False,
                retracted=True,
                purged=False)

            self.assertEqual(pubsub_event, properties.pubsub_event)

        event = '''
            <message from='test@test.test' id='b5ac48d0-0f9c-11dc-8754-001143d5d5db'>
                <items node='princely_musings'>
                    <retract id='ae890ac52d0df67ed7cfdf51b644e901'/>
                </items>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16))

        self.dispatcher.process_data(event)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.const import Chatstate
from nbxmpp.modules.base import BaseModule


class Chatstates(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message_chatstate,
                          ns=Namespace.CHATSTATES,
                          priority=15),
        ]

    def _process_message_chatstate(self, _client, stanza, properties):
        try:
            chatstate = parse_chatstate(stanza)
        except ValueError as error:
            self._log.warning('Invalid chatstate: %s', error)
            self._log.warning(stanza)
            return

        if chatstate is None:
            return

        if properties.is_mam_message:
            return

        properties.chatstate = chatstate


def parse_chatstate(stanza):
    children = stanza.getChildren()
    for child in children:
        if child.getNamespace() == Namespace.CHATSTATES:
            return Chatstate(child.getName())
    return None



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import BookmarkData
from nbxmpp.structs import PubSubEventData


class BookmarkTest(StanzaHandlerTest):

    def test_bookmark_1_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = [
                BookmarkData(jid=JID.from_string('theplay@conference.shakespeare.lit'),
                             name='The Play\'s the Thing',
                             autojoin=True,
                             password='pass',
                             nick='JC'),
                BookmarkData(jid=JID.from_string('second@conference.shakespeare.lit'),
                             name='Second room',
                             autojoin=False,
                             password=None,
                             nick=None)
            ]

            pubsub_event = PubSubEventData(
                node='storage:bookmarks',
                id='current',
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False)

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = '''
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='storage:bookmarks'>
                        <item id='current'>
                            <storage xmlns='storage:bookmarks'>
                                <conference name='The Play&apos;s the Thing'
                                            autojoin='true'
                                            jid='theplay@conference.shakespeare.lit'>
                                    <password>pass</password>
                                    <nick>JC</nick>
                                </conference>
                                <conference name='Second room'
                                            autojoin='0'
                                            jid='second@conference.shakespeare.lit'>
                                </conference>
                            </storage>
                        </item>
                    </items>
                </event>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT))

        self.dispatcher.process_data(event)

    def test_bookmark_2_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = BookmarkData(jid=JID.from_string('theplay@conference.shakespeare.lit'),
                                name='The Play\'s the Thing',
                                autojoin=True,
                                password=None,
                                nick='JC')

            pubsub_event = PubSubEventData(
                node='urn:xmpp:bookmarks:0',
                id='theplay@conference.shakespeare.lit',
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False)

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = '''
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='urn:xmpp:bookmarks:0'>
                        <item id='theplay@conference.shakespeare.lit'>
                            <conference xmlns='urn:xmpp:bookmarks:0'
                                        name='The Play&apos;s the Thing'
                                        autojoin='1'>
                                <nick>JC</nick>
                            </conference>
                        </item>
                    </items>
                </event>
            </message>
        '''

        self.dispatcher.register_handler(
            StanzaHandler(name='message',
                          callback=_on_message,
                          ns=Namespace.PUBSUB_EVENT))

        self.dispatcher.process_data(event)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.task import iq_request_task
from nbxmpp.errors import StanzaError
from nbxmpp.modules.util import process_response
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.bookmarks.util import build_storage_node
from nbxmpp.modules.bookmarks.util import get_private_request
from nbxmpp.modules.bookmarks.util import parse_private_bookmarks



class PrivateBookmarks(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_bookmarks(self):
        _task = yield

        response = yield get_private_request()
        if response.isError():
            raise StanzaError(response)

        bookmarks = parse_private_bookmarks(response, self._log)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        yield bookmarks

    @iq_request_task
    def store_bookmarks(self, bookmarks):
        _task = yield

        self._log.info('Store Bookmarks')

        storage_node = build_storage_node(bookmarks)
        response = yield Iq('set', Namespace.PRIVATE, payload=storage_node)
        yield process_response(response)



from __future__ import annotations

from typing import Callable

import functools
import re


# https://www.rfc-editor.org/rfc/rfc3987
ucschar = (
    '\xA0-\uD7FF'
    '\uF900-\uFDCF'
    '\uFDF0-\uFFEF'
    '\U00010000-\U0001FFFD'
    '\U00020000-\U0002FFFD'
    '\U00030000-\U0003FFFD'
    '\U00040000-\U0004FFFD'
    '\U00050000-\U0005FFFD'
    '\U00060000-\U0006FFFD'
    '\U00070000-\U0007FFFD'
    '\U00080000-\U0008FFFD'
    '\U00090000-\U0009FFFD'
    '\U000A0000-\U000AFFFD'
    '\U000B0000-\U000BFFFD'
    '\U000C0000-\U000CFFFD'
    '\U000D0000-\U000DFFFD'
    '\U000E1000-\U000EFFFD'
)

ALPHA           = 'A-Za-z'
DIGIT           = '0-9'
unreserved      = fr'{ALPHA}{DIGIT}\-\._\~'
subdelims       = "!$&'()*+,;="
iunreserved     = f'{unreserved}{ucschar}'
ipchar          = f'{iunreserved}{re.escape(subdelims)}:@'
ifragment       = fr'{ipchar}/\?'

# https://www.rfc-editor.org/rfc/rfc5122.html#section-2.2
nodeallow       = r"!$()*+,;="
resallow        = r"!$&'()*+,:;="
inode           = f'{iunreserved}{re.escape(nodeallow)}'
ires            = f'{iunreserved}{re.escape(resallow)}'
ivalue          = f'{iunreserved}'

rx_iunreserved  = re.compile(f'[{iunreserved}]*')
rx_inode        = re.compile(f'[{inode}]')
rx_ires         = re.compile(f'[{ires}]')
rx_ikey         = rx_iunreserved
rx_iquerytype   = rx_iunreserved
rx_ivalue       = rx_iunreserved
rx_ifragment     = re.compile(f'[{ifragment}]')


class _Quoter(dict[str, str]):
    """A mapping from a string to its percent encoded form.

    Mapping is only done if string is not in safe range.

    Keeps a cache internally, via __missing__, for efficiency (lookups
    of cached keys don't call Python code at all).
    """
    def __init__(self, safe: re.Pattern[str]) -> None:
        self._safe = safe

    def __repr__(self):
        return f"<Quoter {dict(self)!r}>"

    def __missing__(self, b: str):
        if len(b) != 1:
            raise ValueError("String must be exactly one character long")

        if self._safe.fullmatch(b) is None:
            res = "".join(['%{:02X}'.format(i) for i in b.encode()])
        else:
            res = b
        self[b] = res
        return res


@functools.lru_cache
def _quoter_factory(safe: re.Pattern[str]) -> Callable[[str], str]:
    return _Quoter(safe).__getitem__


def validate_ikey(ikey: str) -> str:
    res = rx_ikey.fullmatch(ikey)
    if res is None:
        raise ValueError('Not allowed characters in key')
    return ikey


def validate_querytype(querytype: str) -> str:
    res = rx_iquerytype.fullmatch(querytype)
    if res is None:
        raise ValueError('Not allowed characters in querytype')
    return querytype


def _escape(string: str, pattern: re.Pattern[str]) -> str:
    quoter = _quoter_factory(safe=pattern)
    return ''.join([quoter(c) for c in string])


def escape_ifragment(fragment: str) -> str:
    return _escape(fragment, rx_ifragment)


def escape_ivalue(value: str) -> str:
    return _escape(value, rx_ivalue)


def escape_inode(node: str) -> str:
    return _escape(node, rx_inode)


def escape_ires(res: str) -> str:
    return _escape(res, rx_ires)


def clean_iri(iri_str: str) -> str:
    if not iri_str.startswith('xmpp:'):
        raise ValueError('IRI must start with xmpp scheme')

    iri_str = iri_str.removeprefix('xmpp:')

    if iri_str.startswith('//'):
        # Remove auth component
        iri_str = iri_str.removeprefix('//')
        iri_str = iri_str.split('/', maxsplit=1)[1]

    # Remove query and fragment
    iri_str = iri_str.split('?', maxsplit=1)[0]
    iri_str = iri_str.split('#', maxsplit=1)[0]
    return iri_str



from __future__ import annotations

from typing import Any

from gi.repository import GLib

import os
import unittest
from unittest.mock import Mock
from nbxmpp.client import Client
from nbxmpp.http import HTTPSession
from nbxmpp.structs import MuclumbusResult
from nbxmpp.modules.muclumbus import Muclumbus

from nbxmpp.sasl import SCRAM_SHA_1
from nbxmpp.util import b64encode

# Test vector from https://wiki.xmpp.org/web/SASL_and_SCRAM-SHA-1

API_URL = 'https://search.jabber.network/api/1.0/search'


@unittest.skipUnless(os.environ.get('NBXMPP_EXTERNAL_UNIT_TESTS'),
                     'ENV var for external tests not set')
class TestMuclumbus(unittest.TestCase):
    def setUp(self):
        self._client = Client()
        self._client.set_http_session(HTTPSession())
        self._module = Muclumbus(self._client)

    def test_http_request(self):

        mainloop = GLib.MainLoop()

        def _result(task: Any) -> None:
            result = task.finish()
            assert isinstance(result, MuclumbusResult)
            assert len(result.items) > 0
            mainloop.quit()

        self._module.set_http_search(API_URL, 'gajim', callback=_result)

        mainloop.run()

if __name__ == '__main__':
    unittest.main()



'''
Module with dummy classes for unit testing of XMPP and related code.
'''

import threading, time

from unittest.mock import Mock

from nbxmpp import idlequeue

IDLEQUEUE_INTERVAL = 0.2 # polling interval. 200ms is used in Gajim as default
IDLEMOCK_TIMEOUT = 30 # how long we wait for an event

class IdleQueueThread(threading.Thread):
    '''
    Thread for regular processing of idlequeue.
    '''
    def __init__(self):
        self.iq = idlequeue.SelectIdleQueue()
        self.stop = threading.Event() # Event to stop the thread main loop.
        self.stop.clear()
        threading.Thread.__init__(self)

    def run(self):
        while not self.stop.isSet():
            self.iq.process()
            time.sleep(IDLEQUEUE_INTERVAL)

    def stop_thread(self):
        self.stop.set()


class IdleMock:
    '''
    Serves as template for testing objects that are normally controlled by GUI.
    Allows to wait for asynchronous callbacks with wait() method.
    '''
    def __init__(self):
        self._event = threading.Event()
        self._event.clear()

    def wait(self):
        '''
        Block until some callback sets the event and clearing the event
        subsequently.
        Returns True if event was set, False on timeout
        '''
        self._event.wait(IDLEMOCK_TIMEOUT)
        if self._event.isSet():
            self._event.clear()
            return True
        else:
            return False

    def set_event(self):
        self._event.set()


class MockConnection(IdleMock, Mock):
    '''
    Class simulating Connection class from src/common/connection.py

    It is derived from Mock in order to avoid defining all methods
    from real Connection that are called from NBClient or Dispatcher
    ( _event_dispatcher for example)
    '''

    def __init__(self, *args):
        self.connect_succeeded = True
        IdleMock.__init__(self)
        Mock.__init__(self, *args)

    def on_connect(self, success, *args):
        '''
        Method called after connecting - after receiving <stream:features>
        from server (NOT after TLS stream restart) or connect failure
        '''
        self.connect_succeeded = success
        self.set_event()


    def on_auth(self, con, auth):
        '''
        Method called after authentication, regardless of the result.

        :Parameters:
                con : NonBlockingClient
                        reference to authenticated object
                auth : string
                        type of authetication in case of success ('old_auth', 'sasl') or
                        None in case of auth failure
        '''
        self.auth_connection = con
        self.auth = auth
        self.set_event()



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.protocol import Error as ErrorStanza
from nbxmpp.protocol import ERR_BAD_REQUEST
from nbxmpp.protocol import ERR_FEATURE_NOT_IMPLEMENTED
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Iq
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import IBBData
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.task import iq_request_task


class IBB(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._process_ibb,
                          ns=Namespace.IBB,
                          priority=20),
        ]

    def _process_ibb(self, _client, stanza, properties):
        if properties.type.is_set:
            open_ = stanza.getTag('open', namespace=Namespace.IBB)
            if open_ is not None:
                properties.ibb = self._parse_open(stanza, open_)
                return

            close = stanza.getTag('close', namespace=Namespace.IBB)
            if close is not None:
                properties.ibb = self._parse_close(stanza, close)
                return

            data = stanza.getTag('data', namespace=Namespace.IBB)
            if data is not None:
                properties.ibb = self._parse_data(stanza, data)
                return

    def _parse_open(self, stanza, open_):
        attrs = open_.getAttrs()
        try:
            block_size = int(attrs.get('block-size'))
        except Exception as error:
            self._log.warning(error)
            self._log.warning(stanza)
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        if block_size > 65535:
            self._log.warning('Invalid block-size')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        sid = attrs.get('sid')
        if not sid:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        type_ = attrs.get('stanza')
        if type_ == 'message':
            self._client.send_stanza(ErrorStanza(stanza,
                                                 ERR_FEATURE_NOT_IMPLEMENTED))
            raise NodeProcessed

        return IBBData(type='open', block_size=block_size, sid=sid)

    def _parse_close(self, stanza, close):
        sid = close.getAttrs().get('sid')
        if sid is None:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed
        return IBBData(type='close', sid=sid)

    def _parse_data(self, stanza, data):
        attrs = data.getAttrs()

        sid = attrs.get('sid')
        if sid is None:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            seq = int(attrs.get('seq'))
        except Exception:
            self._log.exception('Invalid seq')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            decoded_data = b64decode(data.getData())
        except Exception:
            self._log.exception('Failed to decode IBB data')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        return IBBData(type='data', sid=sid, seq=seq, data=decoded_data)

    def send_reply(self, stanza, error=None):
        if error is None:
            reply = stanza.buildReply('result')
            reply.getChildren().clear()
        else:
            reply = ErrorStanza(stanza, error)
        self._client.send_stanza(reply)

    @iq_request_task
    def send_open(self, jid, sid, block_size):
        _task = yield

        response = yield _make_ibb_open(jid, sid, block_size)
        yield process_response(response)

    @iq_request_task
    def send_close(self, jid, sid):
        _task = yield

        response = yield _make_ibb_close(jid, sid)
        yield process_response(response)

    @iq_request_task
    def send_data(self, jid, sid, seq, data):
        _task = yield

        response = yield _make_ibb_data(jid, sid, seq, data)
        yield process_response(response)


def _make_ibb_open(jid, sid, block_size):
    iq = Iq('set', to=jid)
    iq.addChild('open',
                {'block-size': block_size, 'sid': sid, 'stanza': 'iq'},
                namespace=Namespace.IBB)
    return iq


def _make_ibb_close(jid, sid):
    iq = Iq('set', to=jid)
    iq.addChild('close', {'sid': sid}, namespace=Namespace.IBB)
    return iq


def _make_ibb_data(jid, sid, seq, data):
    iq = Iq('set', to=jid)
    ibb_data = iq.addChild('data',
                           {'sid': sid, 'seq': seq},
                           namespace=Namespace.IBB)
    ibb_data.setData(b64encode(data))
    return iq



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Any
from typing import Optional

import time
import logging

from nbxmpp.protocol import Iq
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ErrorNode
from nbxmpp.protocol import ERR_ITEM_NOT_FOUND
from nbxmpp.protocol import NodeProcessed
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.base import BaseModule
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import DiscoItems
from nbxmpp.structs import DiscoItem
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError


log = logging.getLogger('nbxmpp.m.discovery')


class Discovery(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._process_disco_info,
                          typ='get',
                          ns=Namespace.DISCO_INFO,
                          priority=90),
        ]

    @staticmethod
    def _process_disco_info(client, stanza, _properties):
        iq = stanza.buildReply('error')
        iq.addChild(node=ErrorNode(ERR_ITEM_NOT_FOUND))
        client.send_stanza(iq)
        raise NodeProcessed

    @iq_request_task
    def disco_info(self, jid, node=None):
        _task = yield

        response = yield get_disco_request(Namespace.DISCO_INFO, jid, node)
        if response.isError():
            raise StanzaError(response)
        yield parse_disco_info(response)

    @iq_request_task
    def disco_items(self, jid, node=None):
        _task = yield

        response = yield get_disco_request(Namespace.DISCO_ITEMS, jid, node)
        if response.isError():
            raise StanzaError(response)
        yield parse_disco_items(response)



def parse_disco_info(stanza: Any,
                     timestamp: Optional[float] = None) -> DiscoInfo:
    idenities = []
    features = []
    dataforms = []

    if timestamp is None:
        timestamp = time.time()

    query = stanza.getQuery()
    for node in query.getTags('identity'):
        attrs = node.getAttrs()
        try:
            idenities.append(
                DiscoIdentity(category=attrs['category'],
                              type=attrs['type'],
                              name=attrs.get('name'),
                              lang=attrs.get('xml:lang')))
        except Exception:
            raise MalformedStanzaError('invalid attributes', stanza)

    for node in query.getTags('feature'):
        try:
            features.append(node.getAttr('var'))
        except Exception:
            raise MalformedStanzaError('invalid attributes', stanza)

    for node in query.getTags('x', namespace=Namespace.DATA):
        dataforms.append(extend_form(node))

    return DiscoInfo(stanza=stanza,
                     identities=idenities,
                     features=features,
                     dataforms=dataforms,
                     timestamp=timestamp)


def parse_disco_items(stanza):
    items = []

    query = stanza.getQuery()
    for node in query.getTags('item'):
        attrs = node.getAttrs()
        try:
            items.append(
                DiscoItem(jid=attrs['jid'],
                          name=attrs.get('name'),
                          node=attrs.get('node')))
        except Exception:
            raise MalformedStanzaError('invalid attributes', stanza)

    return DiscoItems(jid=stanza.getFrom(),
                      node=query.getAttr('node'),
                      items=items)


def get_disco_request(namespace, jid, node=None):
    iq = Iq('get', to=jid, queryNS=namespace)
    if node:
        iq.getQuery().setAttr('node', node)
    return iq



import tempfile
import unittest
from unittest.mock import Mock
import os
from pathlib import Path

from gi.repository import GLib
from gi.repository import Soup

from nbxmpp.const import HTTPRequestError
from nbxmpp.http import HTTPSession
from test.lib.util import raise_all_exceptions


SMALL_FILE_URL = 'https://gajim.org/downloads/ci/unittest_small_file'  # 200 KB
BIG_FILE_URL = 'https://gajim.org/downloads/ci/unittest_big_file'      # 7   MB
LARGE_FILE_URL = 'https://gajim.org/downloads/ci/unittest_large_file'  # 80  MB
NO_FILE_URL = 'https://gajim.org/downloads/ci/no-file'


# import logging
# consoleloghandler = logging.StreamHandler()
# log = logging.getLogger('nbxmpp.http')
# log.setLevel(logging.DEBUG)
# log.addHandler(consoleloghandler)


@unittest.skipUnless(os.environ.get('NBXMPP_EXTERNAL_UNIT_TESTS'),
                     'ENV var for external tests not set')
class HTTP(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def test_download_to_file(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()

        temp_dir = tempfile.gettempdir()
        request.set_response_body_from_path(Path(temp_dir) / 'file')
        request.connect('response-progress', callback_mock.progress)
        request.connect('starting-response-body', callback_mock.starting)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())

        request.send('GET', SMALL_FILE_URL, timeout=10)

        mainloop.run()

        self.assertTrue(request.is_complete())
        self.assertTrue(request.is_finished())
        self.assertIsNone(request.get_error())

        callback_mock.progress.assert_called()
        callback_mock.starting.assert_called()
        callback_mock.finished.assert_called()

    def test_in_memory_download(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()
        request.connect('response-progress', callback_mock.progress)
        request.connect('starting-response-body', callback_mock.starting)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', SMALL_FILE_URL, timeout=10)

        mainloop.run()

        self.assertTrue(request.is_complete())
        self.assertTrue(request.is_finished())
        self.assertIsNone(request.get_error())

        callback_mock.progress.assert_called()
        callback_mock.starting.assert_called()
        callback_mock.finished.assert_called()

    def test_download_cancelled_before_start(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()
        request.connect('response-progress', callback_mock.progress)
        request.connect('starting-response-body', callback_mock.starting)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', SMALL_FILE_URL, timeout=10)

        GLib.timeout_add(10, request.cancel)

        mainloop.run()

        self.assertFalse(request.is_complete())
        self.assertTrue(request.is_finished())
        self.assertEqual(request.get_error(), HTTPRequestError.CANCELLED)

        callback_mock.progress.assert_not_called()
        callback_mock.starting.assert_not_called()
        callback_mock.finished.assert_called()

    def test_download_cancelled_after_start(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        def _on_start(req):
            req.cancel()

        callback_mock = Mock()
        request.connect('starting-response-body', _on_start)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', LARGE_FILE_URL, timeout=10)

        mainloop.run()

        self.assertFalse(request.is_complete())
        self.assertTrue(request.is_finished())
        self.assertEqual(request.get_error(), HTTPRequestError.CANCELLED)

        callback_mock.finished.assert_called()

    def test_download_failed_404(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()
        request.connect('finished', callback_mock.finished)
        request.connect('response-progress', callback_mock.progress)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', NO_FILE_URL, timeout=5)

        mainloop.run()

        self.assertFalse(request.is_complete())
        self.assertTrue(request.is_finished())
        self.assertEqual(request.get_error(), HTTPRequestError.STATUS_NOT_OK)
        self.assertEqual(request.get_status(), Soup.Status.NOT_FOUND)

        callback_mock.progress.assert_not_called()
        callback_mock.finished.assert_called()

    def test_cancel_with_timeout(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()
        request.connect('starting-response-body', callback_mock.starting)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', LARGE_FILE_URL, timeout=1)

        mainloop.run()

        self.assertTrue(request.is_finished())
        self.assertFalse(request.is_complete())
        self.assertEqual(request.get_error(), HTTPRequestError.TIMEOUT)

        callback_mock.starting.assert_called()
        callback_mock.finished.assert_called()

    def test_signal_propagation(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        callback_mock = Mock()
        request.connect('starting', callback_mock.starting, 1, 2, 3)
        request.connect('got-body', callback_mock.got_body)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', SMALL_FILE_URL, timeout=10)

        mainloop.run()

        self.assertTrue(request.is_finished())
        self.assertTrue(request.is_complete())

        callback_mock.starting.assert_called_with(request, 1, 2, 3)
        callback_mock.got_body.assert_called_with(request)

    def test_parallel_download(self):

        mainloop = GLib.MainLoop()

        request1 = HTTPSession().create_request()
        request1.send('GET', SMALL_FILE_URL, timeout=5)

        request2 = HTTPSession().create_request()
        request2.send('GET', SMALL_FILE_URL, timeout=5)

        request3 = HTTPSession().create_request()
        request3.send('GET', SMALL_FILE_URL, timeout=5)

        request4 = HTTPSession().create_request()
        request4.send('GET', SMALL_FILE_URL, timeout=5)

        GLib.timeout_add_seconds(5, mainloop.quit)

        mainloop.run()

        self.assertTrue(request1.is_finished())
        self.assertTrue(request1.is_complete())

        self.assertTrue(request2.is_finished())
        self.assertTrue(request2.is_complete())

        self.assertTrue(request3.is_finished())
        self.assertTrue(request3.is_complete())

        self.assertTrue(request4.is_finished())
        self.assertTrue(request4.is_complete())

    @raise_all_exceptions
    def test_content_overflow(self):

        mainloop = GLib.MainLoop()

        session = HTTPSession()
        request = session.create_request()

        def _on_starting(req) -> None:
            req._received_size = 100000000000

        callback_mock = Mock()
        request.connect('starting-response-body', _on_starting)
        request.connect('finished', callback_mock.finished)
        request.connect('destroy', lambda *args: mainloop.quit())
        request.send('GET', SMALL_FILE_URL, timeout=10)

        mainloop.run()

        self.assertTrue(request.is_finished())
        self.assertFalse(request.is_complete())
        self.assertEqual(request.get_error(), HTTPRequestError.CONTENT_OVERFLOW)

        callback_mock.finished.assert_called()

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging
import functools
import inspect
from urllib.parse import urlparse
from urllib.parse import unquote

from nbxmpp.structs import CommonResult
from nbxmpp.errors import StanzaError
from nbxmpp.errors import is_error
from nbxmpp.simplexml import Node


def process_response(response):
    if response.isError():
        raise StanzaError(response)

    return CommonResult(jid=response.getFrom())


def raise_if_error(result):
    if is_error(result):
        raise result


def finalize(task, result):
    if is_error(result):
        raise result
    if isinstance(result, Node):
        return task.set_result(result)
    return result


def parse_xmpp_uri(uri):
    url = urlparse(uri)
    if url.scheme != 'xmpp':
        raise ValueError('not a xmpp uri')

    if not ';' in url.query:
        return (url.path, url.query, {})

    action, query = url.query.split(';', 1)
    key_value_pairs = query.split(';')

    dict_ = {}
    for key_value in key_value_pairs:
        key, value = key_value.split('=')
        dict_[key] = unquote(value)

    return (url.path, action, dict_)


def make_func_arguments_string(func, self, args, kwargs):
    signature = inspect.signature(func)
    bound_arguments = signature.bind(self, *args, **kwargs)
    bound_arguments.apply_defaults()
    arg_string = ''
    for name, arg in bound_arguments.arguments.items():
        if name == 'self':
            continue
        arg_string += f'{name}={arg}, '
    arg_string = arg_string[:-2]
    return f'{func.__name__}({arg_string})'


def log_calls(func):
    @functools.wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if self._log.isEnabledFor(logging.INFO):
            self._log.info(make_func_arguments_string(func, self, args, kwargs))
        return func(self, *args, **kwargs)
    return func_wrapper



# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

# XEP-0425: Message Moderation

from typing import Optional

from nbxmpp import JID
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import ModerationData
from nbxmpp.simplexml import Node
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.util import process_response


class Moderation(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message,
                          typ='groupchat',
                          ns=Namespace.FASTEN,
                          priority=20),
            StanzaHandler(name='message',
                          callback=self._process_message_moderated_tombstone,
                          typ='groupchat',
                          ns=Namespace.MESSAGE_MODERATE,
                          priority=20),
        ]

    @iq_request_task
    def send_retract_request(self, muc_jid: JID, stanza_id: str,
                             reason: Optional[str] = None):
        _task = yield

        response = yield _make_retract_request(muc_jid, stanza_id, reason)

        yield process_response(response)

    def _process_message_moderated_tombstone(
        self,
        _client,
        stanza: Node,
        properties: MessageProperties
    ) -> None:

        if not properties.is_mam_message:
            return

        moderated = stanza.getTag(
            'moderated', namespace=Namespace.MESSAGE_MODERATE)
        if moderated is None:
            return

        properties.moderation = self._parse_moderated(
            moderated, properties, properties.mam.id)

    def _process_message(self,
                         _client,
                         stanza: Node,
                         properties: MessageProperties) -> None:

        if not properties.jid.is_bare:
            return

        apply_to = stanza.getTag(
            'apply-to', namespace=Namespace.FASTEN)
        if apply_to is None:
            return

        moderated = apply_to.getTag(
            'moderated', namespace=Namespace.MESSAGE_MODERATE)
        if moderated is None:
            return

        stanza_id = apply_to.getAttr('id')
        if stanza_id is None:
            self._log.warning('apply-to element without stanza-id')
            self._log.warning(stanza)
            raise NodeProcessed

        properties.moderation = self._parse_moderated(moderated, properties, stanza_id)

    def _parse_moderated(
        self,
        moderated: Node,
        properties: MessageProperties,
        stanza_id: str
    ) -> ModerationData | None:

        retract, is_tombstone = _get_retract_element(moderated, properties)
        if retract is None:
            self._log.warning('Failed to find <retract/ed> element')
            return None

        try:
            by = _parse_by_attr(moderated)
        except ValueError as error:
            self._log.warning(error)
            by = None

        stamp = _parse_moderation_timestamp(retract, is_tombstone, properties)

        occupant_id = moderated.getTagAttr('occupant-id',
                                           'id',
                                           namespace=Namespace.OCCUPANT_ID)

        return ModerationData(
            stanza_id=stanza_id,
            moderator_jid=str(by),
            by=by,
            occupant_id=occupant_id,
            reason=moderated.getTagData('reason'),
            timestamp=retract.getAttr('stamp'),
            stamp=stamp,
            is_tombstone=is_tombstone)


def _get_retract_element(
    moderated: Node,
    properties: MessageProperties
) -> tuple[Node | None, bool]:

    retract = moderated.getTag(
            'retract', namespace=Namespace.MESSAGE_RETRACT)
    if retract is not None:
        return retract, False

    retracted = moderated.getTag(
        'retracted', namespace=Namespace.MESSAGE_RETRACT)
    if retracted is not None:
        return retracted, True and properties.is_mam_message
    return None, False


def _parse_by_attr(moderated: Node) -> JID | None:
    by_attr = moderated.getAttr('by')
    if by_attr is None:
        return None

    try:
        return JID.from_string(by_attr)
    except Exception as error:
        raise ValueError('Invalid JID: %s, %s' % (by_attr, error))


def _parse_moderation_timestamp(
    retract: Node,
    is_tombstone: bool,
    properties: MessageProperties
) -> float:

    if is_tombstone:
        stamp_attr = retract.getAttr('stamp')
        stamp = parse_datetime(
            stamp_attr, check_utc=True, convert='utc', epoch=True)
        if stamp is not None:
            return stamp

    if properties.is_mam_message:
        return properties.mam.timestamp

    return properties.timestamp


def _make_retract_request(muc_jid: JID, stanza_id: str,
                          reason: Optional[str]) -> Iq:
    iq = Iq('set', Namespace.FASTEN, to=muc_jid)
    query = iq.setQuery(name='apply-to')
    query.setAttr('id', stanza_id)
    moderate = query.addChild(name='moderate',
                              namespace=Namespace.MESSAGE_MODERATE)
    moderate.addChild(name='retract', namespace=Namespace.MESSAGE_RETRACT)
    if reason is not None:
        moderate.addChild(name='reason', payload=[reason])
    return iq



import os
import unittest

from nbxmpp.protocol import LocalpartByteLimit
from nbxmpp.protocol import LocalpartNotAllowedChar
from nbxmpp.protocol import ResourcepartByteLimit
from nbxmpp.protocol import ResourcepartNotAllowedChar
from nbxmpp.protocol import DomainpartByteLimit
from nbxmpp.protocol import DomainpartNotAllowedChar
from nbxmpp.protocol import JID


class JIDParsing(unittest.TestCase):

    def test_valid_jids(self):
        tests = [
            'juliet@example.com',
            'juliet@example.com/foo',
            'juliet@example.com/foo bar',
            'juliet@example.com/foo@bar',
            'foo\\20bar@example.com',
            'fussball@example.com',
            'fu\U000000DFball@example.com',
            '\U000003C0@example.com',
            '\U000003A3@example.com/foo',
            '\U000003C3@example.com/foo',
            '\U000003C2@example.com/foo',
            'king@example.com/\U0000265A',
            'example.com',
            'example.com/foobar',
            'a.example.com/b@example.net',
        ]

        for jid in tests:
            JID.from_string(jid)

    def test_invalid_jids(self):
        tests = [
            ('"juliet"@example.com', LocalpartNotAllowedChar),
            ('foo bar@example.com', LocalpartNotAllowedChar),
            ('@example.com', LocalpartByteLimit),
            ('user@example.com/', ResourcepartByteLimit),
            ('user@example.com/\U00000001', ResourcepartNotAllowedChar),
            ('user@host@example.com', DomainpartNotAllowedChar),
            ('juliet@', DomainpartByteLimit),
            ('/foobar', DomainpartByteLimit),
        ]

        for jid, exception in tests:
            with self.assertRaises(exception):
                JID.from_string(jid)

    def test_invalid_precis_jids(self):
        os.environ['NBXMPP_ENFORCE_PRECIS'] = 'true'
        tests = [
            ('henry\U00002163@example.com', LocalpartNotAllowedChar),
            ('\U0000265A@example.com', LocalpartNotAllowedChar),
        ]

        for jid, exception in tests:
            with self.assertRaises(exception):
                JID.from_string(jid)

        del os.environ['NBXMPP_ENFORCE_PRECIS']

    def test_ip_literals(self):
        tests = [
            ('juliet@[2002:4559:1FE2::4559:1FE2]/res'),
            ('juliet@123.123.123.123/res'),
        ]

        for jid in tests:
            JID.from_string(jid)

    def test_jid_equality(self):
        tests = [
            'juliet@example.com',
            'juliet@example.com/foo',
            'example.com',
        ]

        for jid in tests:
            self.assertTrue(JID.from_string(jid) == JID.from_string(jid))

    def test_jid_escaping(self):
        # (user input, escaped)
        tests = [
            (r'space cadet@example.com',
             r'space\20cadet@example.com'),

            (r'call me "ishmael"@example.com',
             r'call\20me\20\22ishmael\22@example.com'),

            (r'at&t guy@example.com',
             r'at\26t\20guy@example.com'),

            ('d\'artagnan@example.com',
             r'd\27artagnan@example.com'),

            (r'/.fanboy@example.com',
             r'\2f.fanboy@example.com'),

            (r'::foo::@example.com',
             r'\3a\3afoo\3a\3a@example.com'),

            (r'<foo>@example.com',
             r'\3cfoo\3e@example.com'),

            (r'user@host@example.com',
             r'user\40host@example.com'),

            (r'c:\net@example.com',
             r'c\3a\net@example.com'),

            (r'c:\\net@example.com',
             r'c\3a\\net@example.com'),

            (r'c:\cool stuff@example.com',
             r'c\3a\cool\20stuff@example.com'),

            (r'c:\5commas@example.com',
             r'c\3a\5c5commas@example.com'),

            (r'call me\20@example.com',
             r'call\20me\5c20@example.com'),
        ]

        tests2 = [
            'juliet@example.com',
            'juliet@example.com',
            'juliet@example.com',
            'juliet@example.com',
            'fussball@example.com',
            'fu\U000000DFball@example.com',
            '\U000003C0@example.com',
            '\U000003A3@example.com',
            '\U000003C3@example.com',
            '\U000003C2@example.com',
            'example.com',
        ]

        test3 = [
            '\\20callme\\20@example.com',
            '\\20callme@example.com',
            'callme\\20@example.com',
        ]

        test4 = [
            ('call\\20me@example.com', 'call me@example.com',)
        ]

        fail_tests = [
            r'c:\5commas@example.com/asd',
            r'juliet@example.com/test'
        ]

        for user_input, escaped in tests:
            # Parse user input and escape it
            jid = JID.from_user_input(user_input)

            self.assertTrue(jid.domain == 'example.com')
            self.assertTrue(str(jid) == escaped)
            self.assertTrue(jid.to_user_string() == user_input)

            # We must fail on invalid JIDs
            with self.assertRaises(Exception):
                JID.from_string(user_input)

            # Parse escaped JIDs
            jid = JID.from_string(escaped)
            self.assertTrue(str(jid) == escaped)
            self.assertTrue(jid.domain == 'example.com')

        for jid in tests2:
            # show that from_string() and from_user_input() produce the same
            # result for valid bare JIDs
            self.assertTrue(JID.from_string(jid) == JID.from_user_input(jid))

        for jid in test3:
            # JIDs starting or ending with \20 are not escaped
            self.assertTrue(JID.from_string(jid).to_user_string() == jid)

        for user_input, user_string in test4:
            # Test escaped keyword argument
            self.assertTrue(JID.from_user_input(user_input, escaped=True) != JID.from_user_input(user_input))
            self.assertTrue(JID.from_user_input(user_input, escaped=True).to_user_string() == user_string)

        for user_input in fail_tests:
            # from_user_input does only support bare jids
            with self.assertRaises(Exception):
                JID.from_user_input(user_input)

    def test_jid_to_iri(self):
        tests = [
            ('nasty!#$%()*+,-.;=?[\\]^_`{|}~node@example.com', 'xmpp:nasty!%23$%25()*+,-.;=%3F%5B%5C%5D%5E_%60%7B%7C%7D~node@example.com'),
            ('node@example.com/repulsive !#"$%&\'()*+,-./:;<=>?@[\\]^_`{|}~resource', 'xmpp:node@example.com/repulsive%20!%23%22$%25&\'()*+,-.%2F:;%3C=%3E%3F%40%5B%5C%5D%5E_%60%7B%7C%7D~resource'),
            ('jiři@čechy.example/v Praze', 'xmpp:jiři@čechy.example/v%20Praze')
        ]

        for jid, iri in tests:
            jid = JID.from_string(jid)
            self.assertEqual(jid.to_iri(), iri)

        jid = JID.from_string('example-node@example.com')
        iri = jid.to_iri(('message', [('subject', 'Hello World')]), 'frag')
        self.assertEqual(iri, 'xmpp:example-node@example.com?message;subject=Hello%20World#frag')

        iri = jid.to_iri('message')
        self.assertEqual(iri, 'xmpp:example-node@example.com?message')

        iri = jid.to_iri(fragment='onlyfragment')
        self.assertEqual(iri, 'xmpp:example-node@example.com#onlyfragment')

        jid = JID.from_user_input('call me "ishmael"@example.com')
        self.assertEqual(jid.to_iri(), 'xmpp:call%5C20me%5C20%5C22ishmael%5C22@example.com')

    def test_iri_to_jid(self):
        tests = [
            ('nasty!#$%()*+,-.;=?[\\]^_`{|}~node@example.com', 'xmpp:nasty!%23$%25()*+,-.;=%3F%5B%5C%5D%5E_%60%7B%7C%7D~node@example.com'),
            ('node@example.com/repulsive !#"$%&\'()*+,-./:;<=>?@[\\]^_`{|}~resource', 'xmpp:node@example.com/repulsive%20!%23%22$%25&\'()*+,-.%2F:;%3C=%3E%3F%40%5B%5C%5D%5E_%60%7B%7C%7D~resource'),
            ('jiři@čechy.example/v Praze', 'xmpp:jiři@čechy.example/v%20Praze')
        ]

        for jid_string, iri_string in tests:
            parsed_jid = JID.from_iri(iri_string)
            self.assertEqual(str(parsed_jid), jid_string)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import HTTPAuthData
from nbxmpp.modules.base import BaseModule


class HTTPAuth(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)
        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_http_auth,
                          ns=Namespace.HTTP_AUTH,
                          priority=40),
            StanzaHandler(name='iq',
                          callback=self._process_http_auth,
                          typ='get',
                          ns=Namespace.HTTP_AUTH,
                          priority=40)
        ]

    def _process_http_auth(self, _client, stanza, properties):
        confirm = stanza.getTag('confirm', namespace=Namespace.HTTP_AUTH)
        if confirm is None:
            return

        attrs = confirm.getAttrs()
        body = stanza.getTagData('body')
        id_ = attrs.get('id')
        method = attrs.get('method')
        url = attrs.get('url')
        properties.http_auth = HTTPAuthData(id_, method, url, body)
        self._log.info('HTTPAuth received: %s %s %s %s',
                       id_, method, url, body)



#!/usr/bin/env python3

import sys
import subprocess

DISABLED_CHECKS = [
    'C0103',  # invalid-name
    'C0201',  # consider-iterating-dictionary
    'C0206',  # consider-using-f-string
    'C0209',  # consider-using-dict-items
    'C0301',  # line-too-long
    'C0326',  # bad-whitespace
    'C0330',  # bad-continuation
    'W0201',  # attribute-defined-outside-init
    'W0212',  # protected-access
    'W0221',  # arguments-differ
    'W0223',  # abstract-method
    'W0231',  # super-init-not-called
    'W0233',  # non-parent-init-called
    'W0621',  # redefined-outer-name
    'W0622',  # redefined-builtin
    'W0707',  # raise-missing-from
    'R0201',  # no-self-use
    'R0801',  # duplicate-code
    'E1101',  # no-member
    'E1135',  # unsupported-membership-test
]


def run_pylint_test():

    cmd = [
        'pylint',
        'nbxmpp',
        f'--disable={",".join(DISABLED_CHECKS)}',
        '--ignore=examples'
    ]

    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError:
        sys.exit('pylint test failed')


if __name__ == '__main__':
    run_pylint_test()



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import namedtuple

from nbxmpp.http import HTTPRequest
from nbxmpp.http import HTTPSession
from nbxmpp.util import Observable
from nbxmpp.const import ConnectionType
from nbxmpp.const import ConnectionProtocol
from nbxmpp.util import parse_websocket_uri


log = logging.getLogger('nbxmpp.addresses')


class ServerAddress(namedtuple('ServerAddress', 'domain service host uri '
                                                'protocol type proxy')):

    __slots__ = []

    @property
    def is_service(self):
        return self.service is not None

    @property
    def is_host(self):
        return self.host is not None

    @property
    def is_uri(self):
        return self.uri is not None

    def has_proxy(self):
        return self.proxy is not None


class ServerAddresses(Observable):
    '''
    Signals:

        resolved

    '''

    def __init__(self, domain):
        Observable.__init__(self, log)

        self._domain = domain
        self._http_session = None
        self._custom_host = None
        self._proxy = None
        self._is_resolved = False

        self._addresses = [
            ServerAddress(domain=self._domain,
                          service='xmpps-client',
                          host=None,
                          uri=None,
                          protocol=ConnectionProtocol.TCP,
                          type=ConnectionType.DIRECT_TLS,
                          proxy=None),

            ServerAddress(domain=self._domain,
                          service='xmpp-client',
                          host=None,
                          uri=None,
                          protocol=ConnectionProtocol.TCP,
                          type=ConnectionType.START_TLS,
                          proxy=None),

            ServerAddress(domain=self._domain,
                          service='xmpp-client',
                          host=None,
                          uri=None,
                          protocol=ConnectionProtocol.TCP,
                          type=ConnectionType.PLAIN,
                          proxy=None)
        ]

        self._fallback_addresses = [
            ServerAddress(domain=self._domain,
                          service=None,
                          host='%s:%s' % (self._domain, 5222),
                          uri=None,
                          protocol=ConnectionProtocol.TCP,
                          type=ConnectionType.START_TLS,
                          proxy=None),

            ServerAddress(domain=self._domain,
                          service=None,
                          host='%s:%s' % (self._domain, 5222),
                          uri=None,
                          protocol=ConnectionProtocol.TCP,
                          type=ConnectionType.PLAIN,
                          proxy=None)
        ]

    @property
    def domain(self):
        return self._domain

    @property
    def is_resolved(self):
        return self._is_resolved

    def resolve(self):
        if self._is_resolved:
            self._on_request_resolved()
            return

        if self._proxy is not None:
            # Let the proxy resolve the domain
            self._on_request_resolved()
            return

        if self._custom_host is not None:
            self._on_request_resolved()
            return

        self._resolve_alternatives()

    def _resolve_alternatives(self) -> None:
        if self._http_session is None:
            self._on_request_resolved()
            return

        request = self._http_session.create_request()
        request.send('GET', f'https://{self._domain}/.well-known/host-meta',
                     timeout=5,
                     callback=self._on_alternatives_result)

    def _on_alternatives_result(self, request: HTTPRequest) -> None:

        if not request.is_complete():
            log.info('Failed to retrieve host-meta file: %s',
                     request.get_error_string())
            self._on_request_resolved()
            return

        response_body = request.get_data()
        if not response_body:
            log.info('No response body data found')
            self._on_request_resolved()
            return

        try:
            uri = parse_websocket_uri(response_body.decode())
        except Exception as error:
            log.info('Error parsing websocket uri: %s', error)
            self._on_request_resolved()
            return

        if uri.startswith('wss'):
            type_ = ConnectionType.DIRECT_TLS
        elif uri.startswith('ws'):
            type_ = ConnectionType.PLAIN
        else:
            log.warning('Invalid websocket uri: %s', uri)
            self._on_request_resolved()
            return

        addr = ServerAddress(domain=self._domain,
                             service=None,
                             host=None,
                             uri=uri,
                             protocol=ConnectionProtocol.WEBSOCKET,
                             type=type_,
                             proxy=None)
        self._addresses.append(addr)

        self._on_request_resolved()

    def cancel_resolve(self):
        self.remove_subscriptions()

    def set_custom_host(self, address):
        # Set a custom host, overwrites all other addresses
        self._custom_host = address
        if address is None:
            return

        host_or_uri, protocol, type_ = address
        if protocol == ConnectionProtocol.WEBSOCKET:
            host, uri = None, host_or_uri
        else:
            host, uri = host_or_uri, None

        self._fallback_addresses = []
        self._addresses = [
            ServerAddress(domain=self._domain,
                          service=None,
                          host=host,
                          uri=uri,
                          protocol=protocol,
                          type=type_,
                          proxy=None)]

    def set_proxy(self, proxy):
        self._proxy = proxy

    def set_http_session(self, session: HTTPSession) -> None:
        self._http_session = session

    def _on_request_resolved(self):
        self._is_resolved = True
        self.notify('resolved')
        self.remove_subscriptions()

    def get_next_address(self,
                         allowed_types,
                         allowed_protocols):
        '''
        Selects next address
        '''

        for addr in self._filter_allowed(self._addresses,
                                         allowed_types,
                                         allowed_protocols):
            yield self._assure_proxy(addr)

        for addr in self._filter_allowed(self._fallback_addresses,
                                         allowed_types,
                                         allowed_protocols):
            yield self._assure_proxy(addr)

        raise NoMoreAddresses

    def _assure_proxy(self, addr):
        if self._proxy is None:
            return addr

        if addr.protocol == ConnectionProtocol.TCP:
            return addr._replace(proxy=self._proxy)

        return addr

    def _filter_allowed(self, addresses, allowed_types, allowed_protocols):
        if self._proxy is not None:
            addresses = filter(lambda addr: addr.host is not None, addresses)

        addresses = filter(lambda addr: addr.type in allowed_types,
                           addresses)
        addresses = filter(lambda addr: addr.protocol in allowed_protocols,
                           addresses)
        return addresses

    def __str__(self):
        addresses = self._addresses + self._fallback_addresses
        return '\n'.join([str(addr) for addr in addresses])


class NoMoreAddresses(Exception):
    pass
